import io
import re
import zipfile
from collections.abc import Mapping

import openpyxl
import xltpl.writerx as writerx
from docx import Document
from docxtpl import DocxTemplate
from jinja2.exceptions import TemplateSyntaxError
from rest_framework import exceptions

from document_merge_service.api.data import django_file

from . import models
from .jinja import get_jinja_env, get_jinja_filters


class _MagicPlaceholder(str, Mapping):  # type: ignore
    def __new__(cls, parent=None, name=None):
        self = str.__new__(cls, name if name else "")
        self._parent = parent
        self._reports = parent._reports if parent else set()

        if self != "":
            self._reports.add(self)
        return self

    @property
    def reports(self):
        return list(self._reports)

    def __iter__(self):
        return (x for x in [_MagicPlaceholder(parent=self, name=f"{self}[]")])

    def __getitem__(self, idx):
        assert isinstance(idx, str)
        return _MagicPlaceholder(parent=self, name=f"{self}.{idx}".strip("."))

    def __getattr__(self, attr):
        return _MagicPlaceholder(parent=self, name=f"{self}.{attr}".strip("."))

    def __len__(self):
        return 2

    def __radd__(self, other):
        return str(self) + str(other)


class DocxValidator:
    def _validate_is_docx(self):
        try:
            Document(self.template)
        except (ValueError, zipfile.BadZipfile):
            raise exceptions.ParseError("not a valid docx file")
        finally:
            self.template.seek(0)

    def validate_template_syntax(self, available_placeholders=None):  # pragma: no cover
        raise NotImplementedError(
            "validate_template_syntax must be implemented in engine class"
        )

    def validate(self, available_placeholders=None, sample_data=None):
        self._validate_is_docx()
        self.validate_template_syntax(available_placeholders, sample_data)

    def validate_available_placeholders(
        self, used_placeholders, available_placeholders
    ):
        # We don't validate available_placeholders if it's not given
        if not available_placeholders:
            return

        available_placeholders = self._normalize_available_placeholders(
            available_placeholders
        )

        referenced_unavailable = "; ".join(
            sorted(set(used_placeholders) - set(available_placeholders))
        )
        if referenced_unavailable:
            raise exceptions.ValidationError(
                f"Template uses unavailable placeholders: {referenced_unavailable}"
            )

    def _normalize_available_placeholders(self, placeholders):
        available_placeholders = set(placeholders)
        # add all prefixes of placeholders, so users don't
        # have to add "foo" if they have "foo.bar" in the list
        for ph in placeholders:
            prefix = ""
            for word in ph.split("."):
                prefix = f"{prefix}.{word}" if prefix else word
                if prefix.endswith("[]"):
                    available_placeholders.add(prefix[:-2])
                available_placeholders.add(prefix)
        return available_placeholders


class DocxTemplateEngine(DocxValidator):
    def __init__(self, template):
        self.template = template

    def _extract_image_placeholders(self, doc):
        """Extract placeholders using the image filter.

        This method extracts all placeholder names that use the image filter so
        we can add a dummy image to the sample data for validation. We
        explicitly need to parse headers and footers as well as the actual
        document body.
        """

        body_xml = doc.get_xml()
        body_xml = doc.patch_xml(body_xml)

        xmls = [body_xml]

        for _, part in [
            *doc.get_headers_footers(doc.HEADER_URI),
            *doc.get_headers_footers(doc.FOOTER_URI),
        ]:
            part_xml = doc.get_part_xml(part)
            part_xml = doc.patch_xml(part_xml)
            xmls.append(part_xml)

        images = set()

        for xml in xmls:
            image_match = re.match(r".*{{\s?(\S*)\s?\|\s?image\(.*", xml)
            matches = image_match.groups() if image_match else []
            images.update(matches)

        return images

    def validate_template_syntax(self, available_placeholders=None, sample_data=None):
        try:
            doc = DocxTemplate(self.template)
            root = _MagicPlaceholder()
            env = get_jinja_env()
            ph = {
                name: root[name] for name in doc.get_undeclared_template_variables(env)
            }

            for image in self._extract_image_placeholders(doc):
                cleaned_image = image.strip('"').strip("'")
                ph[root[cleaned_image]] = django_file("black.png").file

            ph["_tpl"] = doc

            doc.render(ph, env)

            if sample_data:
                sample_data["_tpl"] = doc
                doc.render(sample_data, env)

            self.validate_available_placeholders(
                used_placeholders=root.reports,
                available_placeholders=available_placeholders,
            )

        except TemplateSyntaxError as exc:
            arg_str = ";".join(exc.args)
            raise exceptions.ValidationError(f"Syntax error in template: {arg_str}")

        except TypeError as exc:
            if "unsupported operand type(s) for -" in exc.args[0]:
                raise exceptions.ValidationError(
                    'Syntax error in template: "-" is not allowed in placeholders'
                )
            else:
                raise exceptions.ValidationError(*exc.args)

        finally:
            self.template.seek(0)

    def merge(self, data, buf):
        doc = DocxTemplate(self.template)
        data["_tpl"] = doc

        doc.render(data, get_jinja_env(), autoescape=True)
        doc.save(buf)
        return buf


_placeholder_match = re.compile(r"^\s*{{\s*([^{}]+)\s*}}\s*$")


class XlsxTemplateEngine:
    BUILTIN_VARS = [
        "tpl_name",
        "sheet_name",
        "[]",
        "sheet_name.decode",
    ]

    def __init__(self, template):
        self.template = template
        self.writer = None

    def validate_is_xlsx(self):
        try:
            openpyxl.load_workbook(self.template)
        except (ValueError, zipfile.BadZipfile):
            raise exceptions.ParseError("not a valid xlsx file")

    def validate(self, available_placeholders=None, sample_data=None):
        self.validate_is_xlsx()
        self.validate_template_syntax(available_placeholders, sample_data)

    def _expand_available_placeholders(self, ph_list):
        """Expand available placeholder list for (internal) correctness.

        If client gives "foo[].bar", we implicitly also allow "foo[]" and "foo":

        >>> self._expand_available_placeholders(["foo[].bar", "baz.boo"])
        ["foo[]", "foo[].bar", "baz.boo"]
        """
        out_list = []
        for ph in ph_list:
            pieces = ph.split(".")
            for offset in range(len(pieces)):
                prefixed = ".".join(pieces[:offset])
                out_list.append(prefixed)
                if prefixed.endswith("[]"):
                    out_list.append(prefixed[:-2])
            out_list.append(ph)
        return out_list

    def validate_template_syntax(self, available_placeholders=None, sample_data=None):
        # We cannot use jinja to validate because xltpl uses jinja's lexer directly
        magic = None
        if not sample_data:
            sample_data = magic = _MagicPlaceholder()
        buf = io.BytesIO()

        try:
            self.merge(sample_data, buf, is_test_merge=True)
        except TemplateSyntaxError as exc:
            arg_str = ";".join(exc.args)
            raise exceptions.ValidationError(f"Syntax error in template: {arg_str}")

        if available_placeholders and magic is not None:
            missing_set = (
                set(magic.reports)
                - set(self._expand_available_placeholders(available_placeholders))
                - set(self.BUILTIN_VARS)
            )
            if not missing_set:
                return

            missing = "; ".join(missing_set)
            raise exceptions.ValidationError(
                f"Placeholders used in template, but not available: {missing}"
            )

    def merge(self, data, buf, is_test_merge=False):
        self.writer = writer = writerx.BookWriter(self.template)
        self._current_data = data

        writer.jinja_env.filters.update(get_jinja_filters())
        if is_test_merge:
            writer.jinja_env.undefined = self._undefined_factory
        writer.jinja_env.globals.update(dir=dir, getattr=getattr)

        payloads = []
        sheets = writer.sheet_resource_map.sheet_state_list
        for sheet in sheets:
            new = dict(data)
            new["sheet_name"] = sheet.name
            new["tpl_name"] = sheet.name
            payloads.append(new)
        writer.render_book(payloads=payloads)
        writer.save(buf)
        return buf

    def _undefined_factory(self, name):
        # For test merges, we set a custom "undefined" factory that
        # doesn't really do undefined, but just fetches the right value
        # from our magic placeholder structure
        return self._current_data[name]


ENGINES = {
    models.Template.DOCX_TEMPLATE: DocxTemplateEngine,
    models.Template.XLSX_TEMPLATE: XlsxTemplateEngine,
}


def get_engine(engine, template):
    return ENGINES[engine](template)



# Generated by Django 4.2.21 on 2025-05-15 08:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0007_template_created_at_template_created_by_group_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='template',
            name='engine',
            field=models.CharField(choices=[('docx-template', 'Docx Template engine (https://github.com/elapouya/python-docx-template)'), ('xlsx-template', 'Xlsx Template engine (https://github.com/zhangyu836/python-xlsx-template)')], max_length=20),
        ),
    ]



import io
import json
import os
import re
import tempfile
import zipfile
from collections import Counter, namedtuple

import openpyxl
import pytest
from django.urls import reverse
from docx import Document
from lxml import etree
from rest_framework import exceptions, status
from syrupy import filters

from document_merge_service.api.authentication import AnonymousUser
from document_merge_service.api.data import django_file

from .. import models, serializers


def get_filename_from_response(response):
    return re.search(r'filename="(.*)"', response["Content-Disposition"])[1]


@pytest.mark.parametrize("template__description", ["test description"])
@pytest.mark.parametrize(
    "query_params,size",
    [
        ({"description__icontains": "test"}, 1),
        ({"description__search": "test"}, 1),
        ({"description__icontains": "unknown"}, 0),
        ({"description__search": "unknown"}, 0),
    ],
)
def test_template_list_query_params(db, admin_client, template, size, query_params):
    url = reverse("template-list")

    response = admin_client.get(url, data=query_params)
    assert response.status_code == status.HTTP_200_OK
    assert response.json()["count"] == size


def test_template_detail(db, client, template, snapshot):
    url = reverse("template-detail", args=[template.pk])

    response = client.get(url)
    assert response.status_code == status.HTTP_200_OK
    assert response.json() == snapshot(
        exclude=filters.props("created_at", "modified_at")
    )


def test_template_download(db, client, template):
    file = django_file("docx-template-syntax.docx")
    template.template.save(os.path.basename(file.name), file)
    template.save()

    url = reverse("template-detail", args=[template.pk])
    response = client.get(url)

    data = response.json()

    download_url = data["template"]

    template_resp = client.get(download_url)

    file.seek(0)
    assert file.read() == template_resp.getvalue()


def test_template_list_with_file(db, client, template):
    file = django_file("docx-template-syntax.docx")
    template.template.save(os.path.basename(file.name), file)
    template.save()

    url = reverse("template-list")
    response = client.get(url)

    assert response.json()["results"][0]["template"] is not None
    assert response.status_code == status.HTTP_200_OK


def test_template_download_url(db, client, template):
    file = django_file("docx-template-syntax.docx")
    template.template.save(os.path.basename(file.name), file)
    template.save()

    serializer = serializers.TemplateSerializer(template)
    field = serializer.fields["template"]
    assert (
        field.to_representation(template)
        == f"/api/v1/template-download/{template.slug}"
    )


@pytest.mark.parametrize(
    "template_name,engine,status_code,require_authentication,authenticated",
    [
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_201_CREATED,
            False,
            False,
        ),
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_401_UNAUTHORIZED,
            True,
            False,
        ),
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
            False,
            False,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_401_UNAUTHORIZED,
            True,
            False,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-template-syntax.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
            True,
            True,
        ),
        (
            "test.txt",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
            False,
            False,
        ),
    ],
)
def test_template_create(
    db,
    client,
    admin_client,
    engine,
    template_name,
    status_code,
    require_authentication,
    settings,
    authenticated,
):
    if authenticated:
        client = admin_client

    settings.REQUIRE_AUTHENTICATION = require_authentication

    url = reverse("template-list")

    template_file = django_file(template_name)
    data = {"slug": "test-slug", "template": template_file.file, "engine": engine}
    response = client.post(url, data=data, format="multipart")

    assert response.status_code == status_code

    if status_code == status.HTTP_201_CREATED:
        data = response.json()
        template_link = data["template"]
        response = client.get(template_link)
        assert response.status_code == status.HTTP_200_OK
        file_ = io.BytesIO(response.getvalue())
        if engine == "xlsx-template":
            openpyxl.load_workbook(file_)
        else:
            Document(file_)


@pytest.mark.parametrize(
    "exists,status_code,method,url",
    [
        (
            True,
            status.HTTP_200_OK,
            "patch",
            reverse("template-detail", args=["foobar"]),
        ),
        (
            False,
            status.HTTP_201_CREATED,
            "post",
            reverse("template-list"),
        ),
    ],
)
@pytest.mark.parametrize("template__slug", ["foobar"])
def test_created_modified(db, admin_client, exists, template, status_code, method, url):
    template_file = django_file("xlsx-template.xlsx")
    data = {
        "slug": "test-slug",
        "template": template_file.file,
        "engine": models.Template.XLSX_TEMPLATE,
    }

    request_method = getattr(admin_client, method)

    if not exists:
        template.delete()

    response = request_method(url, data=data, format="multipart")

    assert response.status_code == status_code
    data = response.json()

    if exists:
        assert data["created_by_user"] is None
        assert data["created_by_group"] is None
    else:
        assert data["created_by_user"] == "admin"
        assert data["created_by_group"] == "admin"

    assert data["modified_by_user"] == "admin"
    assert data["modified_by_group"] == "admin"

    assert data["modified_at"]
    assert data["created_at"]


@pytest.mark.parametrize(
    "status_code, disable_validation",
    [
        (
            status.HTTP_400_BAD_REQUEST,
            "false",
        ),
        (
            status.HTTP_400_BAD_REQUEST,
            "",
        ),
        (
            status.HTTP_201_CREATED,
            "true",
        ),
    ],
)
def test_disable_validation(
    db,
    status_code,
    admin_client,
    settings,
    disable_validation,
):
    settings.REQUIRE_AUTHENTICATION = False
    url = reverse("template-list")

    template_file = django_file("docx-template-syntax.docx")
    data = {
        "slug": "test-slug",
        "template": template_file.file,
        "engine": models.Template.DOCX_TEMPLATE,
    }
    if disable_validation:
        data["disable_template_validation"] = disable_validation

    response = admin_client.post(url, data=data, format="multipart")
    assert response.status_code == status_code

    if status_code == status.HTTP_201_CREATED:
        data = response.json()
        template_link = data["template"]
        response = admin_client.get(template_link)
        assert response.status_code == status.HTTP_200_OK
        Document(io.BytesIO(response.getvalue()))


@pytest.mark.parametrize(
    "template_name,available_placeholders,sample_data,files,expect_missing_placeholders,engine,status_code",
    [
        (
            "docx-template-placeholdercheck.docx",
            ["foo", "bar", "baz"],
            None,
            [],
            [
                "bar.some_attr",
                "black.png",
                "list",
                "list[]",
                "list[].attribute",
            ],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            [
                "foo",
                "bar",
                "baz",
                "bar.some_attr",
                "list[].attribute",
                "black.png",
            ],
            None,
            [],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-template-placeholdercheck.docx",
            [
                "foo",
                "bar",
                "baz",
                "bar.some_attr",
                "list[].attribute",
            ],
            None,
            [],
            ["black.png"],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
                "baz": "1234",
                "list": [{"attribute": "value"}],
            },
            [django_file("black.png").file],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {},
            [django_file("black.png").file],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {},
            [],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
                "baz": "1234",
                "list": [{"attribute": "value"}],
            },
            [],
            ["black.png"],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
            },
            [django_file("black.png").file],
            ["baz", "list", "list[]", "list[].attribute"],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template.docx",
            ["test", "blah"],
            {"test": "hello"},
            [],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "xlsx-template.xlsx",
            [],
            {"test": "hello"},
            [django_file("black.png").file],
            [],
            models.Template.XLSX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-image-placeholder-header-footer.docx",
            ["black.png", "white.png"],
            None,
            [],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
    ],
)
@pytest.mark.parametrize("use_json", [True, False])
def test_template_create_with_available_placeholders(
    db,
    admin_client,
    engine,
    template_name,
    available_placeholders,
    sample_data,
    files,
    status_code,
    settings,
    use_json,
    expect_missing_placeholders,
):
    settings.DOCXTEMPLATE_JINJA_EXTENSIONS = ["jinja2.ext.loopcontrols"]
    url = reverse("template-list")

    template_file = django_file(template_name)

    # files are being reused, so make sure they're readable
    for f in files:
        f.seek(0)

    data = {
        "slug": "test-slug",
        "template": template_file.file,
        "files": files,
        "engine": engine,
    }
    if sample_data:
        data["sample_data"] = json.dumps(sample_data)
    if available_placeholders:
        data["available_placeholders"] = (
            json.dumps(available_placeholders) if use_json else available_placeholders
        )

    response = admin_client.post(url, data=data, format="multipart")
    assert response.status_code == status_code, response.json()

    if status_code == status.HTTP_400_BAD_REQUEST:
        resp = response.json()
        expect_missing_str = "; ".join(expect_missing_placeholders)

        if sample_data and available_placeholders:
            # validation only allows one of these two params
            assert (
                resp["non_field_errors"][0]
                == "Only one of available_placeholders and sample_data is allowed"
            )
        elif engine == models.Template.XLSX_TEMPLATE and files:
            assert (
                resp["non_field_errors"][0]
                == 'Files are only accepted with the "docx-template" engine'
            )
        elif not sample_data and files:
            assert (
                resp["non_field_errors"][0]
                == "Files are only accepted when also providing sample_data"
            )
        else:
            # we expect some missing placeholders
            assert resp["non_field_errors"][0] in [
                f"Template uses unavailable placeholders: {expect_missing_str}",
                f'No file for image "{expect_missing_str}" provided!',
            ]

    if status_code == status.HTTP_201_CREATED:
        data = response.json()
        template_link = data["template"]
        response = admin_client.get(template_link)
        assert response.status_code == status.HTTP_200_OK
        Document(io.BytesIO(response.getvalue()))


@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-filters.docx"))],
)
@pytest.mark.parametrize(
    "template_name,status_code",
    [
        ("docx-template.docx", status.HTTP_200_OK),
        ("test.txt", status.HTTP_400_BAD_REQUEST),
    ],
)
def test_template_update(db, client, template, template_name, status_code):
    url = reverse("template-detail", args=[template.pk])

    template_file = django_file(template_name)
    data = {"description": "Test description", "template": template_file.file}
    response = client.patch(url, data=data, format="multipart")
    assert response.status_code == status_code

    if status_code == status.HTTP_200_OK:
        assert os.path.isfile(template.template.path) is False
        template.refresh_from_db()
        assert template.description == "Test description"


@pytest.mark.parametrize("template__template", [django_file("docx-template.docx")])
def test_template_destroy(db, client, template):
    url = reverse("template-detail", args=[template.pk])

    response = client.delete(url)
    assert response.status_code == status.HTTP_204_NO_CONTENT
    assert os.path.isfile(template.template.path) is False


@pytest.mark.parametrize(
    "template__slug,template__engine,template__template",
    [
        (
            "TestNameTemplate",
            models.Template.DOCX_TEMPLATE,
            django_file("docx-template.docx"),
        ),
    ],
)
def test_template_merge_docx(
    db, client, mock_filefield_name_validation, template, snapshot
):
    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"test": "Test input"}}, format="json")
    assert response.status_code == status.HTTP_200_OK
    assert (
        response.get("content-type")
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )
    assert get_filename_from_response(response) == f"{template.slug}.docx"

    docx = Document(io.BytesIO(response.getvalue()))
    xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
    try:
        snapshot.assert_match(xml)
    except AssertionError:  # pragma: no cover
        with open(f"/tmp/{template.slug}.docx", "wb") as output:
            output.write(response.getvalue())
        print("Template output changed. Check file at %s" % output.name)
        raise


@pytest.mark.parametrize(
    "placeholder,template_content",
    [
        ("{{blah}}", {"blah": "blub"}),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": "foo"},
        ),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": ""},
        ),
        # passed data should be escaped
        ("{{escapeme}}", {"escapeme": "<&>"}),
    ],
)
def test_merge_expression(
    docx_template_with_placeholder, client, snapshot, placeholder, template_content
):
    """Test evaluation of some custom expressions.

    Use this test to try out expressions without creating a new docx template for each
    variant.
    """
    template = docx_template_with_placeholder(placeholder)

    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": template_content}, format="json")
    assert response.status_code == status.HTTP_200_OK

    docx = Document(io.BytesIO(response.getvalue()))
    xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
    try:
        snapshot.assert_match(xml)
    except AssertionError:  # pragma: no cover
        with open(f"/tmp/{template.slug}.docx", "wb") as output:
            output.write(response.getvalue())
        print("Template output changed. Check file at %s" % output.name)
        raise


@pytest.mark.parametrize(
    "placeholder,template_content",
    [
        ("{{blah}}", {"blah": "blub"}),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": "foo"},
        ),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": ""},
        ),
    ],
)
@pytest.mark.parametrize(
    "template__engine",
    [models.Template.DOCX_TEMPLATE],
)
def test_validate_expression(
    docx_template_with_placeholder, placeholder, template_content
):
    """Test validation of templates with custom expressions."""
    template = docx_template_with_placeholder(placeholder)

    Request = namedtuple("Request", ["user"])
    serializer = serializers.TemplateSerializer(
        context={"request": Request(AnonymousUser())}
    )
    serializer.instance = template

    serializer.validate({"data": template_content})


# This needs a strange parametrization. If `unoconv_local` is in a separate
# `parametrize()`, the template filename in the second test will be appended with a
# hash and the test fails
@pytest.mark.parametrize(
    "template__engine,template__template,unoconv_local",
    [
        (models.Template.DOCX_TEMPLATE, django_file("docx-template.docx"), True),
        (models.Template.DOCX_TEMPLATE, django_file("docx-template.docx"), False),
    ],
)
def test_template_merge_as_pdf(
    db, settings, unoconv_local, client, mock_filefield_name_validation, template
):
    url = reverse("template-merge", args=[template.pk])

    response = client.post(
        url, data={"data": {"test": "Test input"}, "convert": "pdf"}, format="json"
    )
    assert response.status_code == status.HTTP_200_OK
    assert response["Content-Type"] == "application/pdf"
    assert get_filename_from_response(response) == f"{template.slug}.pdf"
    assert response.content[0:4] == b"%PDF"


@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-loopcontrols.docx"))],
)
def test_template_merge_jinja_extensions_docx(
    db, client, mock_filefield_name_validation, template, settings, snapshot
):
    settings.DOCXTEMPLATE_JINJA_EXTENSIONS = ["jinja2.ext.loopcontrols"]

    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"test": "Test input"}}, format="json")
    assert response.status_code == status.HTTP_200_OK
    assert (
        response.get("content-type")
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )

    docx = Document(io.BytesIO(response.getvalue()))
    xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
    snapshot.assert_match(xml)


@pytest.mark.parametrize(
    "missing_file,wrong_mime,status_code",
    [
        (False, False, status.HTTP_200_OK),
        (False, True, status.HTTP_400_BAD_REQUEST),
        (True, False, status.HTTP_400_BAD_REQUEST),
    ],
)
@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-filters.docx"))],
)
def test_template_merge_jinja_filters_docx(
    db,
    client,
    mock_filefield_name_validation,
    template,
    snapshot,
    settings,
    tmp_path,
    missing_file,
    wrong_mime,
    status_code,
):
    settings.LANGUAGE_CODE = "de-ch"
    url = reverse("template-merge", args=[template.pk])

    # Couldn't put this into `parametrize`. For some reason, in the second run, the
    # template name is extended with a seemingly random string.
    template.template = django_file("docx-template-filters.docx")
    template.save()

    data = {
        "data": json.dumps(
            {
                "test_date": "1984-09-15",
                "test_time": "23:24",
                "test_datetime": "1984-09-15 23:23",
                "test_datetime2": "23:23-1984-09-15",
                "test_none": None,
                "test_nested": {"multiline": "This is\na test."},
            }
        ),
    }

    if not missing_file:
        file = django_file("black.png").file
        if wrong_mime:
            # create a file with the correct filename (black.png) but with
            # the contents of a docx.
            file = tmp_path / "black.png"
            for line in template.template.file:
                file.write_bytes(line)
            file = file.open("rb")

        data["files"] = [file]

    response = client.post(url, data=data, format="multipart")
    assert response.status_code == status_code

    if status_code == status.HTTP_200_OK:
        assert (
            response.get("content-type")
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        )

        docx = Document(io.BytesIO(response.getvalue()))
        xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
        snapshot.assert_match(xml)


@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-filters.docx"))],
)
@pytest.mark.parametrize(
    "file_value",
    [None, ""],
)
def test_template_merge_file_reset(
    db,
    client,
    mock_filefield_name_validation,
    template,
    settings,
    file_value,
):
    settings.LANGUAGE_CODE = "de-ch"
    url = reverse("template-merge", args=[template.pk])

    # Couldn't put this into `parametrize`. For some reason, in the second run, the
    # template name is extended with a seemingly random string.
    template.template = django_file("docx-template-filters.docx")
    template.save()

    data = {
        "data": {
            "test_date": "1984-09-15",
            "test_time": "23:24",
            "test_datetime": "1984-09-15 23:23",
            "test_datetime2": "23:23-1984-09-15",
            "test_none": None,
            "test_nested": {"multiline": "This is\na test."},
            "black.png": file_value,
        }
    }

    response = client.post(url, data=data, format="json")
    assert response.status_code == status.HTTP_200_OK

    assert (
        response.get("content-type")
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )


@pytest.mark.parametrize(
    "sample,expected",
    [
        ({"foo": {"bar": ["foo", "blah"]}}, ["foo", "foo.bar", "foo.bar[]"]),
        (
            {
                "this": {
                    "is": {
                        "a": [
                            {
                                "list": {
                                    "with": {
                                        "a": ["nested", "object", "and", "a", "list"]
                                    }
                                }
                            }
                        ]
                    }
                }
            },
            [
                "this",
                "this.is",
                "this.is.a",
                "this.is.a[]",
                "this.is.a[].list",
                "this.is.a[].list.with",
                "this.is.a[].list.with.a",
                "this.is.a[].list.with.a[]",
            ],
        ),
    ],
)
def test_sample_to_placeholders(sample, expected):
    ts = serializers.TemplateSerializer()
    assert ts._sample_to_placeholders(sample) == sorted(expected)


@pytest.mark.parametrize(
    "template__engine,template__template",
    [
        (
            models.Template.DOCX_TEMPLATE,
            django_file("docx-template-placeholdercheck.docx"),
        )
    ],
)
def test_template_merge_missing_data(
    db, client, mock_filefield_name_validation, template, settings
):
    settings.DOCXTEMPLATE_JINJA_EXTENSIONS = ["jinja2.ext.loopcontrols"]

    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"blah": "Test input"}}, format="json")

    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert response.json() == [
        "Placeholder from template not found in data: 'bar' is undefined"
    ]


@pytest.mark.parametrize(
    "template_fragment,expected_error",
    [
        (
            "{{E-BAU_NUMBER}}",
            'Syntax error in template: "-" is not allowed in placeholders',
        ),
        (
            "{{E*BAU_NUMBER}}",
            "can't multiply sequence by non-int of type '_MagicPlaceholder'",
        ),
        ("{{E?BAU_NUMBER}}", "Syntax error in template: unexpected char '?' at 1033"),
    ],
)
def test_placeholder_with_unsupported_operand(
    db, client, docx_template_with_placeholder, template_fragment, expected_error
):
    template = docx_template_with_placeholder(template_fragment)

    Request = namedtuple("Request", ["user"])
    serializer = serializers.TemplateSerializer(
        context={"request": Request(AnonymousUser())}
    )
    serializer.instance = template

    with pytest.raises(exceptions.ValidationError) as exc_info:
        serializer.validate({"data": {"E_BAU_NUMBER": 12345}})
    assert exc_info.value.args[0] == expected_error


def test_template_merge_docx_libreoffice_bug(
    db, client, mock_filefield_name_validation, template, snapshot
):
    """Verify a certain docx corruption bug does not occur.

    Certain versions of python-docx and python-docxtemplate cause corruption
    of files that were originally created with LibreOffice. One effect of that
    corruption is a duplicate entry in the document-internal files; there
    are two docProps/core.xml files in the resulting document.
    """
    file = django_file("created_with_libreoffice.docx")
    template.template.save(os.path.basename(file.name), file)
    template.engine = "docx-template"
    template.save()
    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"test": "Test input"}}, format="json")

    with tempfile.NamedTemporaryFile(suffix=".docx") as tmp:
        tmp.write(response.content)
        tmp.seek(0)

        zzz = zipfile.ZipFile(tmp.name)
        name_counter = Counter()
        name_counter.update([f.filename for f in zzz.filelist])

    problematic_names = {
        name: count for name, count in name_counter.most_common() if count > 1
    }

    assert problematic_names == {}, "Duplicate entry in docx file's internal structure"



from django.apps import AppConfig
from django.conf import settings
from django.db.models import TextField
from django.db.models.lookups import IContains


class DefaultConfig(AppConfig):
    name = "document_merge_service.api"

    def ready(self):
        if "sqlite3" in settings.DATABASES["default"]["ENGINE"]:  # pragma: no cover
            TextField.register_lookup(IContains, lookup_name="search")
        mitigate_docxtpl_corruption_bug()


def mitigate_docxtpl_corruption_bug():
    # This is basically monkey-patching this PR:
    # https://github.com/python-openxml/python-docx/pull/1436

    # Hold my beer!
    from docx.opc.constants import RELATIONSHIP_TYPE

    if hasattr(RELATIONSHIP_TYPE, "CORE_PROPERTIES_OFFICEDOCUMENT"):  # pragma: no cover
        raise Exception(
            "The docxtpl mitigation is no longer required, please remove the monkeypatch code"
        )

    RELATIONSHIP_TYPE.CORE_PROPERTIES_OFFICEDOCUMENT = (
        "http://schemas.openxmlformats.org/officedocument/2006/relationships"
        "/metadata/core-properties"
    )

    from docx.opc.package import RT, CorePropertiesPart, OpcPackage, cast

    @property
    def _core_properties_part(self) -> CorePropertiesPart:
        """|CorePropertiesPart| object related to this package.

        Creates a default core properties part if one is not present (not common).
        """
        try:
            return cast(CorePropertiesPart, self.part_related_by(RT.CORE_PROPERTIES))
        except KeyError:
            try:
                office_document_part = self.part_related_by(
                    RT.CORE_PROPERTIES_OFFICEDOCUMENT  # type: ignore
                )
                rel = self.relate_to(
                    office_document_part,
                    RT.CORE_PROPERTIES_OFFICEDOCUMENT,  # type: ignore
                )
                self.rels[rel].reltype = RT.CORE_PROPERTIES
                return cast(CorePropertiesPart, office_document_part)
            except KeyError:
                core_properties_part = CorePropertiesPart.default(self)
                self.relate_to(core_properties_part, RT.CORE_PROPERTIES)
                return core_properties_part

    OpcPackage._core_properties_part = _core_properties_part

    from docx.opc.rel import _Relationship

    @_Relationship.reltype.setter
    def reltype(self, value: str):
        self._reltype = value

    _Relationship.reltype = reltype



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import storages
from django.core.management.base import BaseCommand
from tqdm import tqdm

from document_merge_service.api.models import Template


class Command(BaseCommand):
    help = "Swaps plain text template content to encrypted content"

    def handle(self, *args, **options):
        if not settings.DMS_ENABLE_AT_REST_ENCRYPTION:
            return self.stdout.write(
                self.style.WARNING(
                    "Encryption is not enabled. Skipping encryption of templates."
                )
            )

        failed_templates = []

        # flip between default and encrypted storage to have the correct parameters in the requests
        encrypted_storage = storages.create_storage(settings.STORAGES["default"])
        unencrypted_storage_setting = settings.STORAGES["default"]
        if (
            "OPTIONS" not in unencrypted_storage_setting
            or "object_parameters" not in unencrypted_storage_setting["OPTIONS"]
        ):
            raise ImproperlyConfigured(
                "Encryption is enabled but no object_parameters found in the storage settings."
            )
        del unencrypted_storage_setting["OPTIONS"]["object_parameters"]
        unencrypted_storage = storages.create_storage(unencrypted_storage_setting)

        query = Template.objects.all()
        for template in tqdm(query.iterator(50), total=query.count()):
            # get original template content
            template.template.storage = unencrypted_storage
            try:
                content = template.template.open()

                # overwrite with encrypted content
                template.template.storage = encrypted_storage
                template.template.save(template.template.name, content)
            except Exception as e:
                self.stdout.write(
                    self.style.WARNING(f"Error for template {str(template.pk)}: {e}")
                )
                failed_templates.append(str(template.pk))
                continue

        if failed_templates:
            self.stdout.write(
                self.style.WARNING(f"These templates failed:\n{failed_templates}")
            )
        self.stdout.write(self.style.SUCCESS("Encryption finished"))



# Generated by Django 4.2.15 on 2024-08-19 07:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0006_remove_template_group'),
    ]

    operations = [
        migrations.AddField(
            model_name='template',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name='template',
            name='created_by_group',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='template',
            name='created_by_user',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='template',
            name='modified_at',
            field=models.DateTimeField(auto_now=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name='template',
            name='modified_by_group',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='template',
            name='modified_by_user',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
    ]
